#pragma once
#ifndef EXCHANGE_STRUCTS_H
#define EXCHANGE_STRUCTS_H

#include <cstdint>

namespace Datafox
{

#pragma pack(push, 1)

//! Id der Datei.
enum EnumFileId
{
    FileIdUnknown,
    FileIdMainMenu,                                       //!< Datei für Hauptmenu-Konfigurationsdaten
    FileIdDeviceUpdate                                    //!< Geraetedatei.
};

struct StructFileHeader
{
    uint8_t df[2];      //!< Konstante Kennung "DF" 0x4446
    uint16_t id;        //!< Id des Dateninhaltes, EnumFileId
                        //!< 1 = Datei für Hauptmenues
                        // ...
}; // 4 Byte

//! Id der Struktur.
enum EnumStructId
{
    StructIdHeader,                                        //!< StructStructHeader
    StructIdArray,                                         //!< StructStructArray
    StructIdMainMenuNode,                                  //!< StructMainMenuNode
    StructIdMainMenuColor,                                 //!< StructMainMenuColor
    StructIdMainMenuKeyDef,                                //!< StructMainMenuKeyDef
    StructIdMainMenu,                                      //!< StructMainMenu
    StructIdMainMenuKey,                                   //!< StructMainMenuKey
    StructIdDeviceUpdateNode,                              //!< StructDeviceUpdateNode
    StructIdDeviceUpdateSegmentFirmware,                   //!< StructDeviceUpdateSegmentFirmware
    StructIdDeviceUpdateSegment,                           //!< StructDeviceUpdateSegment
    StructIdMainMenuGallery,                               //!< StructMainMenuGallery
    StructIdMainMenuExtra                                  //!< StructMainMenuExtra
};

struct StructStructHeader
{
    uint16_t id;        //!< Id dieser Struktur, EnumStructId
    uint16_t size;      //!< Größe dieser Struktur, sie stellt auch eine Version dar, da Strukturen nur größer werden können.
}; // 4 Byte

struct StructStructArray
{
    StructStructHeader header;  //!< Strukturheader
    uint16_t count;             //!< Anzahl der folgenden Offsetwerte als unsigned int
                                //!< Die Offsetwerte liegen direkt hinter dieser Struktur und beziehen sich auf die Startadresse dieser Struktur.
    //unsigned int[count]
}; // 6 Byte + dynamisch[ count * sizeof(unsigned int) ]

#pragma pack(pop)

} // namespace Datafox

#endif
