#ifndef STUDIO_SETUP_CLASSES_H
#define STUDIO_SETUP_CLASSES_H

#include <boost/signals2/signal.hpp>

#ifdef DATAFOX_STUDIO_IV
#include "DatafoxMFC/control/tab_control_setup_defines.h"
#else
#define PAGE_NO                       -1
#define PAGE_BLANK                     0
#define PAGE_DEVICE_0                  1
#define PAGE_DEVICE_1                  2
#define PAGE_DEVICE_2                  3
#define PAGE_DEVICE_3                  4
#define PAGE_DEVICE_4                  5
#define PAGE_DEVICE_5                  6
#define PAGE_DEVICE_6                  7
#define PAGE_DEVICE_7                  8
#define PAGE_DEVICE_8                  9
#define PAGE_TABLE_0                  10
#define PAGE_TABLE_FIELD_0            11
#define PAGE_HANDLING_0               12
#define PAGE_HANDLING_1               13
#define PAGE_MENU_0                   14
#define PAGE_MENU_1                   15
#define PAGE_MENU_CHOICE_0            16
#define PAGE_INPUT_CHAIN_0            17
#define PAGE_INPUT_CHAIN_FIELD_0      18
#define PAGE_INPUT_CHAIN_FIELD_1      19
#define PAGE_INPUT_CHAIN_FIELD_2      20
#define PAGE_INPUT_CHAIN_FIELD_3      21
#define PAGE_INPUT_CHAIN_FIELD_4      22
#define PAGE_INPUT_CHAIN_FIELD_5      23
#define PAGE_VIBRATION_SENSOR_0       24
#define PAGE_GPRS_0                   25
#define PAGE_GPS_0                    26
#define PAGE_TIMEBOY_ADAPTION_0       27
#define PAGE_MACHINES_0               28
#define PAGE_DIGITAL_INPUT_0          29
#define PAGE_ANALOG_INPUT_0           30
#define PAGE_TIMER_0                  31
#define PAGE_TIMER_1                  32
#define PAGE_SYSTEMMESSAGE_0          33
#define PAGE_IOMODULE_0               34
#define PAGE_IOMODULE_1               35
#define PAGE_IOMODULE_2               36
#define PAGE_IOMODULE_3               37
#define PAGE_IOMODULE_4               38
#define PAGE_IOMODULE_5               39
#define PAGE_ACCESS_CONTROL2_0        40
#define PAGE_ACCESS_CONTROL2_1        41
#define PAGE_TABLE_LIST_0             42
#define PAGE_TABLE_RECORD_0           43
#define PAGE_ENERGIE_SCHEMES_0        44
#define PAGE_ENERGIE_SCHEMES_1        45
#define PAGE_ENERGIE_SCHEMES_2        46
#define PAGE_ENERGIE_SCHEMES_3        47
#define PAGE_ENERGIE_SCHEMES_4        48
#define PAGE_ENERGIE_SCHEMES_5        49
#define PAGE_DEVICE_9                 50
#define PAGE_DEVICE_10                51
#define PAGE_DEVICE_11                52
#endif

#include "file/file_aes.h"
#include "datafox/communication/module_hardware.h"
#include "document_studio.h"

#include <string>
#include <cstdint>

#include <boost/format.hpp>

#include <vector>
#include <algorithm>

#include <boost/uuid/uuid.hpp>
#include <boost/uuid/uuid_generators.hpp>

#include <boost/assert.hpp>

class EntryBase;
typedef std::vector<EntryBase *> EntryList;

// Vorwards der Roots
class EntryDevice;
class EntryInputChain;
class EntryTableRecord;
class EntryTableList;
class EntryHandling;
class EntrySignalProcessing;
class EntryEventChains;
class EntryIOModule;
class EntryAccessControl2;
class EntryEnergieSchemes;

/**
    Auflistung aller möglicher Entry-Typen, welche von EntryBase abgeleitet sind.
*/
enum class EntryType
{
    Device,                        // Haupteintrag fuer Geraet
    TableRecord,                   // (1 Ebene) Datensatzbeschreibungen
    TableList,                     // (1 Ebene) Listenbeschreibungen
    Table,                         //  (2 Ebene) Tabelle (Datensatzbeschreibung sowie Listenbeschreibung)
    TableField,                    //   (3 Ebene) Tabellenfeld (fuer Datensatzbeschreibung sowie Listenbeschreibung)
    Handling,                      //  (2 Ebene) Bedienung
    Menu,                          //   (ab 3ter Ebene) Menue (Hauptmenue sowie Untermenue)
    MenuChoice,                    //    (ab 4ter Ebene) Menueintrag
    InputChain,                    //     (ab 5ter Ebene) Eingabekette
    InputChainField,               //      (ab 6ter Ebene) Eingabekettenfeld
    SignalProcessing,              // (1 Ebene) Signalverarbeitung - Konfiguration
    DigitalInputs,                 //  (2 Ebene) Digitale Eingänge
    DigitalInput,                  //   (3 Ebene) Digitaler Eingang
    AnalogInputs,                  //  (2 Ebene) Analoge Eingänge
    AnalogInput,                   //   (3 Ebene) Analoger Eingang
    Timers,                        //  (2 Ebene) Timer
    Timer,                         //   (3 Ebene) Timer
    SystemMessages,                //  (2 Ebene) Systemmeldungen
    SystemMessage,                 //   (3 Ebene) Systemmeldung
    EventChains,                   // (1 Ebene) Eingabeketten der Ereignisverarbeitung
    IOModule,                      // (1 Ebene) IOModul - Konfiguration
    AccessControl,                 // (1 Ebene) Zutrittskontrolle (Version 2)
    EnergieSchemes                 // (1 Ebene) Energieverwaltung
};

/////////////////////////////
// Abstrakte Basisklasse
//
// Von dieser Klasse werden alle anderen direkt abgeleitet.

class EntryBase
{
public:
    // Konstruktor
    EntryBase( EntryBase *parent, EntryType type );
    // Der Kopierkonstruktor wird fuer die Drag&Drop Operation benötigt,
    // dadurch ist ein einfaches kopieren der Klassen moeglich.
    EntryBase( const EntryBase &entry );

    // Destruktor
    virtual ~EntryBase();

    ///////////////////////////////////
    // Eintragstyp
    // Dieses ist um die Eintraege zu unterscheiden. (eigenes RTTI)
private:
    EntryType m_type; // Hält die TypeID des Eintrags.
public:
    // Gibt den Typen des Eintrags zurück.
    inline EntryType type() const
    {
        return m_type;
    };
    // Fragt ab ob der Eintrag vom Type ist.
    inline bool isType( EntryType type ) const
    {
        return (type == m_type);
    }

    // Gibt den Typen des Elterneintrags zurück.
    inline EntryType parentType() const
    {
        BOOST_ASSERT( parent() != nullptr );
        return parent()->type();
    }
    // Gibt den Typen des Elterneintrags zurück.
    inline bool isParentType( EntryType type ) const
    {
        BOOST_ASSERT( parent() != nullptr );
        return (type == parent()->type());
    }
    // Gibt den Typen des Großelterneintrags zurück.
    inline EntryType grandParentType() const
    {
        BOOST_ASSERT( grandParent() != nullptr );
        return grandParent()->type();
    }

private:
    boost::uuids::uuid m_uuid;

private:
    EntryList m_children;
    ///////////////////////////////////
    // Schnittstelle zu den Childs
public:
    EntryList::const_iterator constBegin() const
    {
        return m_children.cbegin();
    }
    EntryList::const_iterator constEnd() const
    {
        return m_children.cend();
    }
    EntryList::value_type first()
    {
        return m_children.front();
    }
    const EntryList::value_type first() const
    {
        return m_children.front();
    }
    EntryList::value_type last()
    {
        return m_children.back();
    }
    const EntryList::value_type last() const
    {
        return m_children.back();
    }

    uint32_t size() const
    {
        return static_cast<uint32_t>( m_children.size() );
    }
    const EntryList::value_type at( std::size_t i ) const
    {
        return m_children[i];
    }
    bool contains( const EntryList::value_type &value ) const
    {
        return std::find( m_children.cbegin(), m_children.cend(), value ) != m_children.cend();
    }
    EntryList::value_type takeLast()
    {
        EntryList::value_type entry = m_children.back();
        m_children.pop_back();

        return entry;
    }
    EntryList::value_type takeAt( std::size_t i )
    {
        EntryList::value_type entry = m_children[i];
        m_children.erase( m_children.cbegin() + i );
        return entry;
    }
    bool empty() const
    {
        return m_children.empty();
    }
    void append( const EntryList::value_type &value )
    {
        m_children.push_back( value );
    }
    void prepend( const EntryList::value_type &value )
    {
        m_children.insert( m_children.begin(), value );
    }
    void insert( std::size_t i, const EntryList::value_type &value )
    {
        m_children.insert( m_children.cbegin() + i, value );
    }
    void clear()
    {
        m_children.clear();
    }
    int32_t indexOf( const EntryList::value_type &value, std::size_t from = 0 ) const
    {
        auto iter = std::find( m_children.cbegin(), m_children.cend(), value );
        if ( iter == m_children.cend() || static_cast<uint32_t>(iter - m_children.cbegin()) < from )
        {
            return -1;
        }

        return static_cast<int32_t>( iter - m_children.cbegin() );
    }
    EntryList::value_type value( std::size_t i ) const
    {
        return m_children[i];
    }
    bool removeOne( const EntryList::value_type &value )
    {
        auto iter = std::find( m_children.cbegin(), m_children.cend(), value );
        if ( iter != m_children.cend() )
        {
            m_children.erase( iter );
            return true;
        }

        return false;
    }

    void traverse( std::function<bool(EntryBase &)> lookup )
    {
        if ( !lookup( *this ) )
        {
            // Abbruch des aufsuchens der Childs.
            return;
        }

        // Alle Untereintraege rekursiv abarbeiten!
        auto iterChild = constBegin();
        for ( ; iterChild != constEnd(); ++iterChild )
        {
            (*iterChild)->traverse( lookup );
        }
    }

    /////////////////////////
    // Elterneintrag
protected:
    EntryList::value_type m_parent; // Hält den Parent von diesem Eintrag.

public:
    // Setzt den Parent dieses Eintrags
    void setParent( EntryBase *parent );
    // Gibt den aktuellen Parent dieses Eintrags zurück.
    inline EntryBase *parent() const
    {
        return m_parent;
    }
    // Gibt den nächsten Parent vom angegeben Typ zurück, oder nullptr.
    inline EntryBase *parent( EntryType type ) const
    {
        EntryBase *entry = m_parent;
        while ( entry && !entry->isType( type ) )
        {
            entry = entry->parent();
        }

        return entry;
    }

    // Gibt den Grandparent dieses Eintrags zurück.
    inline EntryBase *grandParent() const
    {
        BOOST_ASSERT( parent() != nullptr );
        return parent()->parent();
    }
    // Gibt den Parent des Eintrags vom type zurück.
    inline EntryBase *grandParent( EntryType type ) const
    {
        BOOST_ASSERT( parent() != nullptr );
        return parent( type )->parent();
    }

    // Ermitteln eines Haupteintrags im Baum
    // Gibt Referenz auf DeviceRoot zurück.
    EntryDevice &root();
    // Gibt Referenz auf Rooteintrag der Tabellen zurück.
    EntryTableRecord &rootTableRecord();
    // Gibt Referenz auf Rooteintrag der Listen zurück.
    EntryTableList &rootTableList();
    // Gibt Referenz auf Rooteintrag der Bedienung zurück.
    EntryHandling &rootHandling();
    // Gibt Referenz auf Rooteintrag der Signalverarbeitung zurück.
    EntrySignalProcessing &rootSignalProcessing();
    // Gibt Referenz auf Rooteintrag der Ereignisketten zurück.
    EntryEventChains &rootEventChains();
    // Gibt Referenz auf Rooteintrag des IOModule zurück.
    EntryIOModule &rootIOModule();
    // Gibt Referenz auf Rooteintrag der ZK 2 zurück.
    EntryAccessControl2 &rootAccessControl();
    // Gibt Referenz auf Rooteintrag der Energieverwaltung zurück.
    EntryEnergieSchemes &rootEnergieSchemes();

public:
    // Operationen //
    // Helper //
    // Ermittelt ob der Eintrag im Baum, unterhalb des aktuellen Eintrags liegt.
    bool isSubChild( EntryBase *pEntry );
    // Gibt eine Liste von Childs zurück, die von einem bestimmten typ sind. (siehe TID_xxx)
    // Hierbei wird der komplette Baum unterhalb dieses Eintrags durchsucht.
    uint32_t collectAllChildsByType( EntryList &childs, EntryType type );

public:
    // Erzeugt einen Childeintrag vom bestimmbaren Typ.
    virtual EntryBase *insertEntry( EntryType typeID, int pos ) = 0;
    virtual EntryBase *cloneEntry() = 0;
    // Löscht das Child von der bestimmbaren nullindizierten Position.
    virtual void deleteEntry( int pos = -1 );
protected:
    // Fügt den Eintrag in die Liste der Childs ein. // Aufruf über InsertEntry
    void insertEntryFinish( EntryBase *entryNew, int pos = -1 );

    //////////////////////////////////////
    // Referenzierung zwischen Einträgen
    // Da durch Drag&Drop verschiebungen von Objekten im Baum vorkommen,
    // sind "harte" nullbasierte Indizes schlecht zu händeln.
    // Jedes Objekt ist im Speicher über seine eindeutige Adresse definiert,
    // diese wird zur Referenzierung herangezogen.
protected:
    //CTypedPtrList<CPtrList, EntryBase*> m_refs;

public:
    // Konvertiert alle Referenzbezüge des Eintrags von Objektzeigern in Indizes.
    // Ist vor der Serialisierung in die Speicherarrays aufzurufen.
    virtual void convertReferenceToIndex()
    {
        auto child = constBegin();
        for ( ; child != constEnd(); ++child )
        {
            (*child)->convertReferenceToIndex();
        }
    }
    // Konvertiert alle Indizes des Eintrags von Indizes in Objektzeiger.
    // Ist nach der Deserialisierung aus den Speicherarrays aufzurufen.
    virtual void convertIndexToReference()
    {
        auto child = constBegin();
        for ( ; child != constEnd(); ++child )
        {
            (*child)->convertIndexToReference();
        }
    }

public:
    virtual void incRefCount()
    {
        ++m_refCount;
    }
    virtual void decRefCount()
    {
        --m_refCount;
    }
    void setRefCount( short val )
    {
        m_refCount = val;
    }
    short refCount()
    {
        return m_refCount;
    }
private:
    short m_refCount; // Hält die Anzahl der Bezüge zu diesem Eintrag.

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    static uint32_t m_flashOffset; // Offset fuer einzutragende Adressen (fuer Extended - Flashbereich)
public:
    // Speichert alle Daten dieses Eintrags in den Speicherarrays und ruft diese Funktion für alle Childs auf.
    virtual uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) = 0;
    // Liest aus den Speicherarrays die Daten für diesen Eintrag und ruft diese Funktion für alle Childs auf.
    virtual uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) = 0;

    // Helper //
    // Verändert die Byte-Reihenfolge des übergebenen Werts. Motorolla und Intel-Format.
    void changeOrderInt32( uint8_t *buf );
    void changeOrderInt16( uint8_t *buf );

    uint16_t endianReverse( uint16_t val );
    uint32_t endianReverse( uint32_t val );

public:
    // Rekursives Löschen des MODIFIED-Flag vom Root an.
    void treeClearModified();
    // Rekursives Prüfen des MODIFIED-Flag vom Root an.
    bool isTreeModified();

protected:
    // Prüfen des MODIFIED-FLag für Eintrag selbst und alle Childs.
    bool isAnyChildModified();
    // Löschen des MODIFIED-Flag für Eintrag selbst und alle Childs.
    void clearAnyChildModified();

private:
    // Prüfung der Eintragsdaten
    virtual void checkMembers()
    {
    };
      // Aufruf in SetParent, um den Childs den Parent zu uebergeben und Members zu pruefen
    virtual void setChildsParent( EntryBase *pEntry );
    // Prüfen der Eintragsdaten aller Childs.
    virtual void checkChildsMembers();

    /////////////////////////
    // Flag's
protected:
    struct
    {
        uint8_t Insert : 1; // Einfuegen von Untereintraegen moeglich?
        uint8_t Delete : 1; // Loeschen dieses Eintrags moeglich?
        uint8_t Copy : 1; // Kopieren des Eintrags moeglich?
        uint8_t Visible : 1; // Sichtbarer Eintrag im TreeCtrl?
        uint8_t Modified : 1; // Aenderungen an Eintragsdaten vorgenommen?
    } m_typeFlags;

public:
    // Einfuegen von Untereintraegen erlaubt.
    inline void setInsertEnabled( bool bEnabled )
    {
        m_typeFlags.Insert = bEnabled;
    }
    inline bool isInsertEnabled() const
    {
        return m_typeFlags.Insert;
    }
    // Loeschen des Eintrags erlaubt.
    inline void setDeleteEnabled( bool bEnabled )
    {
        m_typeFlags.Delete = bEnabled;
    }
    inline bool isDeleteEnabled() const
    {
        return m_typeFlags.Delete;
    }
    // Kopieren des Eintrags erlaubt.
    inline void setCopyEnabled( bool bEnabled )
    {
        m_typeFlags.Copy = bEnabled;
    }
    inline bool isCopyEnabled() const
    {
        return m_typeFlags.Copy;
    }
    // Eintrag in Baum sichtbar.
    inline void setVisible( bool bVisible )
    {
        m_typeFlags.Visible = bVisible;
    }
    inline bool isVisible() const
    {
        return m_typeFlags.Visible;
    }
    // Daten des Eintrages abgeaendert
    inline void setModified( bool bModified = true )
    {
        m_typeFlags.Modified = bModified;
    }
    inline bool isModified() const
    {
        return m_typeFlags.Modified;
    }

    ///////////////////////////////////
    // TreeCtrl
protected:
    void *m_handleItem; // Handle vom zugehoerigen TreeCtrl - Eintrag
    int m_iconIndex;    // Iconnummer des unselektieren TreeCtrl - Eintrags
    int m_iconIndexSelected; // Iconnummer des selektierten TreeCtrl - Eintrags

private:
    std::string m_name; // Name des Eintrags
public:
    // Gibt den Namen des Eintrags zurück.
    const std::string &name() const
    {
        return m_name;
    }
    // Setzt den Namen des Eintrags.
    void setName( const std::string &name )
    {
        if ( name != m_name )
        {
            m_name = name;
            nameChanged( this );
            displayNameChanged( this );
        }
    }

    // Aenderung des Namens melden.
    typedef boost::signals2::signal<void( EntryBase * )> SignalNameChanged;
    SignalNameChanged nameChanged;

private:
    std::string m_displayNameFormat; // Anzeigeformat für Eintragsname im TreeCtrl
public:
    // Gibt das Anzeigefoormat des Eintragsnamens zurück.
    const std::string &displayNameFormat() const
    {
        return m_displayNameFormat;
    }
    // Setzt das Anzeigeformat des Eintragsnamens.
    void setDisplayNameFormat( const std::string &format )
    {
        if ( format != m_displayNameFormat )
        {
            m_displayNameFormat = format;
            displayNameChanged( this );
        }
    }

    // Aenderung des Anzeigenamens melden.
    typedef boost::signals2::signal<void( EntryBase * ) > SignalDisplayNameChanged;
    SignalDisplayNameChanged displayNameChanged;

public:
    // Gibt den Anzeigenamen des Eintrags zurück.
    std::string displayName() const
    {
        if ( m_displayNameFormat.size() == 0 )
        {
            return m_name;
        }

        return boost::str( boost::format( m_displayNameFormat ) % m_name );
    }

public:
    // Setzt den Treeitem-Handle für diesen Eintrag
    void setTreeItemHandle( void *newHandle )
    {
        m_handleItem = newHandle;
    }
    // Gibt den aktuell gesetzten Treeitem-Handle dieses Eintrags zurück.
    void *treeItemHandle()
    {
        return m_handleItem;
    }
    // Ermittelt ab diesem Eintrag abwärts im Baum, den zugehörigen Eintrag zum Treeitem-Handle.
    EntryBase *findTreeItem( void *hSelTreeItem );

    // Setzt die Indizes für die Icons des Eintrags.
    void setIconIndexes( int iconIndex, int iconIndexSelected )
    {
        m_iconIndex = iconIndex;
        m_iconIndexSelected = iconIndexSelected;
    }
    // Gibt den Index des Icons für diesen Eintrag zurück.
    int iconIndex()
    {
        return m_iconIndex;
    }
    // Gibt den Index für das Selektionsicon für diesen Eintrag zurück.
    int iconIndexSelected()
    {
        return m_iconIndexSelected;
    }


/////////////////////////
// TabCtrl
public:
    // Indexliste der zugeordneten Pages (PAGE_NO -> keine)
    std::vector<int> m_pageIndexes;

public:
    // Gibt den zuletzt gewählten Pageindex des Eintrags zurück. (0 .. MAX_ENTRY_TAB)
    virtual int pageIndex()
    {
        return PAGE_NO;
    }
    // Setzt den Pageindex auf den aktuellen.
    virtual void setPageIndex( int index )
    {
        (void)index;
    }


    ///////////////////////////////////
    // Daten des Eintrags
public:
    // Initialisiert alle Daten des Eintrags mit Vorgabewerten.
    virtual void setDefault() = 0;
};


/////////////////////////////
// Klasse fuer Baumwurzeleintrag

class EntryDevice : public EntryBase
{
public:
    // Konstruktor
    EntryDevice( EntryBase *parent = nullptr, EntryType type = EntryType::Device );
    // Destruktor
    ~EntryDevice() override;

    ///////////////////////////////////
    // Childs
public:
    EntryBase *insertEntry( EntryType typeID, int pos ) override;
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;


    /////////////////////////
    // TabCtrl
private:
    static int m_pageIndex;
public:
    int pageIndex();
    void setPageIndex( int index );

    //////////////////////////////////////
    // Referenzierung zwischen Einträgen
public:
    void convertReferenceToIndex() override;
    void convertIndexToReference() override;

public: // Eingabekettenbezuege fuer m_extended.xxx
    std::array<EntryInputChain *, 2> m_refInputChainEnergieOption;

    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;
    void setDefault( SetupType setupType );

    enum enum_device_flag : uint64_t
    {
        PZE_MasterIV = 0x00000001, // PZE-MasterIV
        ZK_MasterIV = 0x00000002, // ZK-MasterIV
        BDE_MasterIV = 0x00000004, // BDE-MasterIV
        Mobil_MasterIV = 0x00000008, // Mobil-MasterIV
        MDE_BoxIV = 0x00000010, // MDE-BoxIV
        AE_MasterIV = 0x00000020, // AE-MasterIV
        Timeboy_MasterIV = 0x00000040, // TimeboyIV
        Exklusiv_LineIV = 0x00000080, // Exklusiv-Line IV
        Flex_MasterIV = 0x00000100, // Flex-MasterIV
        Timeboy_Mobil_PZE = 0x00000200, // Timeboy-Mobil-PZE
        EVO_2_8_Pure = 0x00000400, // EVO 2.8 Pure
        EVO_4_3 = 0x00000800, // EVO 4.3
        IPC_Extension_VARIO_5_7 = 0x00001000, // IPC-Extensions
        IPC_Extension_VARIO_10 = 0x00002000, // IPC-Extension für Vario 10 / Vario 12 / Vario 15
        IPC_Extension_Q7 = 0x00004000, // IPC-Extension für Q7
        IO_Box = 0x00008000, // IO-Box
        KYO_Inloc = 0x00010000, // KYO Inloc
        Mobil_Box = 0x00020000, // Mobil-Box
        Timeboy_TBDock = 0x00040000, // Dockingstation-V2
        Mobil_TBDock = 0x00080000, // Fahrzeugdatenlogger-V2
        KYO_Oneloc = 0x00100000, // KYO Oneloc
        EVO_3_5_Universal = 0x00200000, // EVO 3.5 Universal
        SAMS_OCL_Reader = 0x00400000, // SAMS-OCL-Reader
        EVO_3_5_Pure = 0x00800000, // EVO 3.5 Pure
        EVO_Agera = 0x01000000, // EVO Agera
        KYO_Cenloc = 0x02000000, // KYO Cenloc
        Vario_10 = 0x04000000,
        Vario_12 = 0x08000000,
        Vario_15 = 0x10000000,
        Vario_17 = 0x20000000,
        IPC_EVO_12_1 = 0x40000000,
        IPC_EVO_15 = 0x80000000,
        IPC_EVO_18_5 = 0x100000000,
        IPC_EVO_24 = 0x200000000,
        EVO_Intera_2 = 0x400000000,
        EVO_4_6_FlexKey = 0x800000000,

        // Terminalgeräte der MasterIV-Serie
        MasterIV_Terminals = (PZE_MasterIV | BDE_MasterIV | AE_MasterIV | Exklusiv_LineIV),
        // Timeboygeräte der MasterIV-Serie
        MasterIV_Timeboys = (Timeboy_MasterIV | Timeboy_Mobil_PZE),
        // Boxgeräte
        MasterIV_Boxes = (ZK_MasterIV | Mobil_MasterIV | MDE_BoxIV),
        Evo_Line_Boxes = (IO_Box | KYO_Inloc | Mobil_Box | KYO_Oneloc),
        Boxes = (MasterIV_Boxes | Evo_Line_Boxes),
        // Geräte der MasterIV-Serie
        MasterIV = (MasterIV_Terminals | MasterIV_Timeboys | MasterIV_Boxes),
        // Terminalgeräte der EVO-Line
        EVO_Terminals = (EVO_2_8_Pure | EVO_4_3 | EVO_3_5_Universal | SAMS_OCL_Reader | EVO_Agera),
        // Geräte er IPC-Extension
        IPC_Extensions = (IPC_Extension_VARIO_5_7),
        // Geräte der EVO_Line
        EVO_Line = (Evo_Line_Boxes | EVO_Terminals | IPC_Extensions),
        // Dockingstationen-V2
        TimeboyDockings = (Timeboy_TBDock | Mobil_TBDock),
        // Terminalgeräte der MasterIV-Serie mit verbauter V4-Baugruppe
        MasterIV_Terminals_V4 = (PZE_MasterIV | AE_MasterIV | Flex_MasterIV)
    };

    bool isTypeDevice( enum_device_flag typeDeviceFlags );

public:
    bool isPasswordProtection()
    {
        return m_passwordProtection;
    }
    void setPasswordProtection( bool enabled )
    {
        m_passwordProtection = enabled;
    }
    bool isHidePassword()
    {
        return m_hidePassword;
    }
    void setHidePassword( bool enabled )
    {
        m_hidePassword = enabled;
    }

    const std::string &password()
    {
        return m_password;
    }
    void setPassword( const std::string &password )
    {
        m_password = password;
    }

    bool hasTranponderMultiConfigs( TransponderType transponderType = TransponderType::Undefined );
    uint8_t transponderConfigCount();
    
public:
    StructConfig m_config;
    StructGlobalVar m_globalVar; // Stuktur für Definition der Globalen Variablen
    StructValueGlobalVar m_valueGlobalVar; // Struktur für Defaultwerte der Globalen Variablen.
    StructExtended m_extended;

    StructTransponder m_transponder;
    StructTransponderMifare m_transponderMifare;
    uint16_t m_transponderActiveFlags;

    StructTransponderLegicAdvantHeader m_transponderLegicAdvantHeader;
    StructTransponderLegicAdvant m_transponderLegicAdvant[5];

    StructTransponderMifareDesfireHeader m_transponderMifareDesfireHeader;
    StructTransponderMifareDesfire m_transponderMifareDesfire[5];

    StructTransponderMifarePlusHeader m_transponderMifarePlusHeader;
    StructTransponderMifarePlus m_transponderMifarePlus[5];

    StructTransponderMifareUltralightHeader m_transponderMifareUltralightHeader;
    StructTransponderMifareUltralight m_transponderMifareUltralight[5];

    StructTransponderiCodeHeader m_transponderiCodeHeader;
    StructTransponderiCode m_transponderiCode[5];

    StructTransponderJavaCardHeader m_transponderJavaCardHeader;
    StructTransponderJavaCard m_transponderJavaCard[7];

    StructTransponderMyDHeader m_transponderMyDHeader;
    StructTransponderMyD m_transponderMyD[5];

    StructTransponderISOTagsHeader m_transponderISOTagsHeader;
    StructTransponderISOTags m_transponderISOTags[1];

    StructTransponderETSITagsHeader m_transponderBlueboxIDTronicHeader;
    StructTransponderETSITags m_transponderBlueboxIDTronic[1];

    StructTransponderWiegandClockDataHeader m_transponderWiegandClockDataHeader;
    StructTransponderWiegandClockData m_transponderWiegandClockData[1];

    StructTransponderQ5Header m_transponderQ5Header;
    StructTransponderQ5 m_transponderQ5[5];

    StructTransponderHitag1FeigHeader m_transponderHitag1FeigHeader;
    StructTransponderHitag1Feig m_transponderHitag1Feig[1];

    StructTransponderPreconfiguredReaderHeader m_transponderPreconfiguredReaderHeader;
    StructTransponderPreconfiguredReader m_transponderPreconfiguredReader[1];

    // Passwortschutz
    bool m_passwordProtection;
    bool m_hidePassword;
    std::string m_password;

private:
    // Setuptyp, für welche Geräte dieses Setup gilt.
    SetupType m_setupType;

public:
    SetupType setupType() const;
    void setSetupType( SetupType setupType );

    bool hasSetupType( const std::set<SetupType> &types )
    {
        return types.count( m_setupType );
    }
    bool hasSetupType( SetupType type )
    {
        return (type == m_setupType);
    }

public:
    TransponderType transponderType() const;
    void setTransponderType( TransponderType transponderType );

    bool hasTransponderType( const std::set<TransponderType> &types )
    {
        return types.count( static_cast<TransponderType>(m_config.transponderType & 0x3F) );
    }
    bool hasTransponderType( TransponderType type )
    {
        return (type == static_cast<TransponderType>(m_config.transponderType & 0x3F));
    }
};


/////////////////////////////
// Klasse fuer Menues (mit Eintaegen fuer F-Tasten)

class EntryMenu : public EntryBase
{
public:
    // Konstruktor
    EntryMenu( EntryBase *parent = 0, EntryType type = EntryType::Menu ) : EntryBase( parent, type )
    {
        // Referenzbezug auf Sprungziele
        m_refJumpESC = nullptr;
    }
    // Der Kopierkonstruktor wird fuer die Drag&Drop Operation benötigt,
    // dadurch ist ein einfaches kopieren der Klassen moeglich.
    EntryMenu( const EntryMenu &entry );

    // Destruktor
    ~EntryMenu() override
    {
    }

    ///////////////////////////////////
    // Childs
public:
    EntryBase *insertEntry( EntryType typeID = EntryType::MenuChoice, int pos = -1 ) override;
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;
    void deleteEntry( int pos = -1 ) override;

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;

    //////////////////////////////////////
    // Referenzierung zwischen Einträgen
public:
    void convertReferenceToIndex() override;
    void convertIndexToReference() override;
public:
    EntryBase *m_refJumpESC;   // Sprungziel nach ESC

    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;
public:
    StructOperating m_operating;
    StructMenu m_menu;
};


/////////////////////////////
// Klasse fuer Hauptmenuwurzel

class EntryHandling : public EntryBase
{
public:
    // Konstruktor
    EntryHandling( EntryBase *parent = 0, EntryType type = EntryType::Handling ) : EntryBase( parent, type )
    {
    }
    // Destruktor
    ~EntryHandling() override
    {
    }

    ///////////////////////////////////
    // Childs
public:
    EntryBase *insertEntry( EntryType typeID = EntryType::Menu, int pos = -1 ) override;
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;

    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;
public:
    StructConfigRoot m_configRoot;
    uint8_t m_autoKey[64]; // 16 * 4 Byte fuer Tastenumschaltung
};


/////////////////////////////
// Klasse fuer Datensatzbeschreibungswurzel

class EntryTableRecord : public EntryBase
{
public:
    // Konstruktor
    EntryTableRecord( EntryBase *parent = 0, EntryType type = EntryType::TableRecord ) : EntryBase( parent, type )
    {
    }
    // Destruktor
    ~EntryTableRecord() override
    {
    }

    ///////////////////////////////////
    // Childs
public:
    EntryBase *insertEntry( EntryType typeID = EntryType::Table, int pos = -1 ) override;
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;
    void deleteEntry( int pos = -1 ) override;

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;

    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;
};


/////////////////////////////
// Klasse fuer Listenbeschreibungswurzel

class EntryTableList : public EntryBase
{
public:
    // Konstruktor
    EntryTableList( EntryBase *parent = 0, EntryType type = EntryType::TableList ) : EntryBase( parent, type )
    {
    }
    // Destruktor
    ~EntryTableList() override
    {
    }

    ///////////////////////////////////
    // Childs
public:
    EntryBase *insertEntry( EntryType typeID = EntryType::Table, int pos = -1 ) override;
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;
    void deleteEntry( int pos = -1 ) override;

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;

    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;
};

/////////////////////////////
// Klasse fuer Datensatz-/Listenbeschreibung

// Childklasse fuer Referenz bekanntgeben
class EntryTableField;

class EntryTable : public EntryBase
{
public:
    // Konstruktor
    EntryTable( EntryBase *parent = 0, EntryType type = EntryType::Table ) : EntryBase( parent, type )
    {
        // Referenz auf Schluesselfeld
        m_refKeyField = nullptr;
        // Referenz auf Prioritaetsfeld
        m_refPriorityField = nullptr;
    }
    // Destruktor
    ~EntryTable() override
    {
    }

    ///////////////////////////////////
    // Childs
public:
    EntryBase *insertEntry( EntryType typeID = EntryType::TableField, int pos = -1 );
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;
    void deleteEntry( int pos = -1 ) override;

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;
private:
    void checkMembers() override;

    //////////////////////////////////////
    // Referenzierung zwischen Einträgen
public:
    void convertReferenceToIndex() override;
    void convertIndexToReference() override;

public:
    EntryTableField *m_refKeyField;      // Zugeordnetes Feld fuer die Sortierspalte von Listen.
    EntryTableField *m_refPriorityField; // Zugeordnetes Feld fuer die Prioritaetsspalte von Datensaetzen.

    ///////////////////////////////////
    // Helper
public:
    // Ermittelt die Gesamtlänge des Datensatzes
    int recordSize();

    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;
    void setDefault( const std::string &name );
public:
    StructTable m_table;
};


/////////////////////////////
// Klasse fuer 

class EntryTableField : public EntryBase
{
public:
    // Konstruktor
    EntryTableField( EntryBase *parent = 0, EntryType type = EntryType::TableField ) : EntryBase( parent, type )
    {
    }
    // Destruktor
    ~EntryTableField() override
    {
    }

    ///////////////////////////////////
    // Childs
public:
    EntryBase *insertEntry( EntryType typeID, int pos ) override;
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;
private:
    void checkMembers() override;

    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;
    void setDefault( const std::string &strName, unsigned char ucType, unsigned char ucLength );
public:
    StructTableField m_tableField;
};

/////////////////////////////
// Klasse fuer Menueeintraege

class EntryMenuChoice : public EntryBase
{
public:
    // Konstruktor
    EntryMenuChoice( EntryBase *parent = 0, EntryType type = EntryType::MenuChoice ) : EntryBase( parent, type )
    {
    }
    // Der Kopierkonstruktor wird fuer die Drag&Drop Operation benötigt,
    // dadurch ist ein einfaches kopieren der Klassen moeglich.
    EntryMenuChoice( const EntryMenuChoice &entry );
    // Destruktor
    ~EntryMenuChoice() override
    {
    }

    ///////////////////////////////////
    // Childs
public:
    EntryBase *insertEntry( EntryType typeID = EntryType::InputChain, int pos = -1 ) override;
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;

    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;
    void setDefault( const std::string &name );
public:
    StructMenuItem m_menuItem;
};


/////////////////////////////
// Klasse fuer Eingabekettenfelder

class EntryInputChainField : public EntryBase
{
public:
    // Konstruktor
    EntryInputChainField( EntryBase *parent = 0, EntryType type = EntryType::InputChainField ) : EntryBase( parent, type )
    {
        // Referenzbezug auf Tabellenfeld initialisieren
        m_refTableField = nullptr;
        m_refTableFieldIntegral = nullptr;
        // Referenzbezug auf Sprungziele
        m_refJumpESC = nullptr;
        m_refJump = nullptr;
        m_refJumpSelNoEntry = nullptr;
        // Referenz fuer Listenzugriff
        m_refList = nullptr;        // Liste aus der ausgewaehlt wird
        m_refAuswahlFeld = nullptr; // Feld aus dem ausgewaehlt wird
        m_refDatFeld = nullptr;     // Feld das abgespeichert wird
        m_refSelDFeld = nullptr;    // Erstes Selektionsfeld der Datensatzbeschreibung
        m_refSelLFeld = nullptr;    // Erstes Selektionsfeld der Liste
        m_refSelDFeld2 = nullptr;   // Zweites Selektionsfeld der Datensatzbeschreibung
        m_refSelLFeld2 = nullptr;   // Zweites Selektionsfeld der Liste
        // Referenz auf Datensatzfeld für den Status der FeverControl-Funktion
        m_refFeverControlStatusField = nullptr;
        // Referenz auf Datensatzfeld für die Kennung der MultiInputSource-Funktion
        m_refMultiInputSourceField = nullptr;

        // Kennung bei einer Multifeldfunktion.
        setMultiInputChildType( MultiInputChildType::None );
    }
    // Der Kopierkonstruktor wird fuer die Drag&Drop Operation benötigt,
    // dadurch ist ein einfaches kopieren der Klassen moeglich.
    EntryInputChainField( const EntryInputChainField &entry );
    // Destruktor
    ~EntryInputChainField() override
    {
        // Referenzbezug auf Tabellenfeld entfernen
        if ( m_refTableField )
            m_refTableField->decRefCount();

        if ( m_refTableFieldIntegral )
            m_refTableFieldIntegral->decRefCount();

        // Referenz fuer Listenzugriff
        if ( m_refList )
        {
            // Liste aus der ausgewaehlt wird

            m_refList->decRefCount();

            // Feld aus dem ausgewaehlt wird
            if ( m_refAuswahlFeld )
                m_refAuswahlFeld->decRefCount();

            // Feld das abgespeichert wird
            if ( m_refDatFeld )
                m_refDatFeld->decRefCount();

            // Erstes Selektionsfeld der Datensatzbeschreibung
            if ( isValidSelRef( m_refSelDFeld ) )
                m_refSelDFeld->decRefCount();

            // Erstes Selektionsfeld der Liste
            if ( m_refSelLFeld )
                m_refSelLFeld->decRefCount();

            // Zweites Selektionsfeld der Datensatzbeschreibung
            if ( isValidSelRef( m_refSelDFeld2 ) )
                m_refSelDFeld2->decRefCount();

            // Zweites Selektionsfeld der Liste
            if ( m_refSelLFeld2 )
                m_refSelLFeld2->decRefCount();
        }

        // Referenz auf Datensatzfeld für den Status der FeverControl-Funktion
        if ( m_refFeverControlStatusField )
        {
            m_refFeverControlStatusField->decRefCount();
        }

        // Referenz auf Datensatzfeld für die Kennung der MultiInputSource-Funktion
        if ( m_refMultiInputSourceField )
        {
            m_refMultiInputSourceField->decRefCount();
        }
    }

    ///////////////////////////////////
    // Childs
public:
    EntryBase *insertEntry( EntryType typeID, int pos ) override;
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;

    enum class MultiInputChildType : uint8_t
    {
        None,
        Keypad,
        Barcode,
        Tranponder,
        FingerprintIdentification,
        FingerprintVerification
    };
protected:
    MultiInputChildType m_multiInputChildType;

public:
    void setMultiInputChildType( MultiInputChildType type );
    MultiInputChildType multiInputChildType();

    EntryInputChainField *insertMultiInputChild( MultiInputChildType type );
    void removeMultiInputChild( MultiInputChildType type );

    EntryInputChainField *multiInputChild( MultiInputChildType type );

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;
private:
    void checkMembers() override;
public:
    void setFieldFunction( uint8_t function );
    uint8_t fieldFunction();

    ///////////////////////////////////
    // PageCtrl
private:
    static int m_pageIndex;
public:
    int pageIndex();
    void setPageIndex( int index );

    //////////////////////////////////////
    // Referenzierung zwischen Einträgen
public:
    void convertReferenceToIndex() override;
    void convertIndexToReference() override;
    // Die ersten 0 - 255 Werte der Entry-Addresse der zugeordneten Selektionsfelder,
    // werden für einen Index einer globalen Variablen genutzt.
    inline bool isValidSelRef( EntryBase *entry )
    {
        return (entry > (EntryBase *)0xFF);
    }

public:
    EntryTableField *m_refTableField{};
    EntryTableField *m_refTableFieldIntegral{};
    EntryBase *m_refJumpESC{}; // Sprungziel bei ESC
    EntryBase *m_refJump{};    // Sprungziel bei -Bedingtem Sprung-
    EntryBase *m_refJumpSelNoEntry{}; // Sprungziel bei -Kein Eintrag in Liste-
    // Referenzen fuer Listenauswahl
    EntryTable *m_refList{};             // Liste aus der ausgewaehlt wird
    EntryTableField *m_refAuswahlFeld{}; // Feld aus dem ausgewaehlt wird
    EntryTableField *m_refDatFeld{};     // Feld das abgespeichert wird
    EntryTableField *m_refSelDFeld{};    // Erstes Selektionsfeld der Datensatzbeschreibung
    EntryTableField *m_refSelDFeld2{};   // Zweites Selektionsfeld der Datensatzbeschreibung
    EntryTableField *m_refSelLFeld{};    // Erstes Selektionsfeld der Liste
    EntryTableField *m_refSelLFeld2{};   // Zweites Selektionsfeld der Liste
    // Referenz auf Datensatzfeld für den Status der FeverControl-Funktion
    EntryTableField *m_refFeverControlStatusField{};
    // Referenz auf Datensatzfeld für die Kennung der MultiInputSource-Funktion
    EntryTableField *m_refMultiInputSourceField;

    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;
    void setDefault( const std::string &name );

public:
    StructTextEx                m_textEx;
    StructConfirmationEx        m_confirmationEx;
    StructNumFieldEx            m_numFieldEx;
    StructListSelectionEx       m_listSelectionEx;
    StructDataEx                m_dataEx;
    StructGlobVarToHIDEx        m_globVarToHIDEx;
    StructSwitchModuleEnergieEx m_switchModuleEnergieEx;
    StructSystemMessageEx       m_systemMessageEx;
    StructTakePhotoEx           m_takePhotoEx;
    StructFeverControlEx        m_feverControlEx;
    StructMultiInputSourceEx    m_multiInputSourceEx;
    StructReadField             m_readField;
};


/////////////////////////////
// Klasse fuer Eingabeketten

class EntryInputChain : public EntryBase
{
public:
    // Konstruktor
    EntryInputChain( EntryBase *parent = 0, EntryType type = EntryType::InputChain ) : EntryBase( parent, type )
    {
        // Referenzbezug auf Tabelle initialisieren
        m_refTable = nullptr;
        // Referenzbezug auf Sprungziele
        m_refJumpCR = nullptr;
        m_refJumpESC = nullptr;
    }
    // Der Kopierkonstruktor wird fuer die Drag&Drop Operation benötigt,
    // dadurch ist ein einfaches kopieren der Klassen moeglich.
    EntryInputChain( const EntryInputChain &entry );
    // Destruktor
    ~EntryInputChain() override
    {
        // Referenzbezug auf Tabelle entfernen
        if ( m_refTable )
        {

            m_refTable->decRefCount();
        }
    }

    ///////////////////////////////////
    // Childs
public:
    EntryBase *insertEntry( EntryType typeID = EntryType::InputChainField, int pos = -1 ) override;
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;
private:
    void checkMembers() override;

    //////////////////////////////////////
    // Referenzierung zwischen Einträgen
public:
    void convertReferenceToIndex() override;
    void convertIndexToReference() override;
public:
    EntryTable *m_refTable;    // Zugeordnete Tabelle
    EntryBase *m_refJumpCR;    // Sprungziel nach Enter
    EntryBase *m_refJumpESC;   // Sprungziel nach ESC

    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;
    void setDefault( const std::string &name, bool visible = true );
public:
    StructReadRecord m_readRecord;
private:
    StructOperating m_operating;
};

/////////////////////////////
// Klasse fuer 

class EntryIOModule : public EntryBase
{
public:
    // Konstruktor
    EntryIOModule( EntryBase *parent = 0, EntryType type = EntryType::IOModule ) : EntryBase( parent, type )
    {
    }
    // Destruktor
    ~EntryIOModule() override
    {
    }

    ///////////////////////////////////
    // Childs
public:
    EntryBase *insertEntry( EntryType typeID, int pos ) override;
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;

    //////////////////////////////////////
    // Referenzierung zwischen Einträgen
public:
    void convertReferenceToIndex() override;
    void convertIndexToReference() override;
public:
    EntryInputChain *m_refInputChainDigital[4][8];


    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;
    void defaultIOModule( StructIOModule *pIOModule );
public:
    StructIOModule m_ioModule;
    StructDigitalInput m_digitalInput[8];
};

/////////////////////////////
// Klasse fuer Digitale Eingänge

class EntryDigialInputs : public EntryBase
{
public:
    // Konstruktor
    EntryDigialInputs( EntryBase *parent = 0, EntryType type = EntryType::DigitalInputs ) : EntryBase( parent, type )
    {
    }
    // Destruktor
    ~EntryDigialInputs() override
    {
    }

    ///////////////////////////////////
    // Childs
public:
    EntryBase *insertEntry( EntryType typeID = EntryType::DigitalInput, int pos = -1 ) override;
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;
    void deleteEntry( int pos = -1 ) override;

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;

    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;

    ////////////////////////////////////
    // Schnittstelle des Eintrags
    StructDigitalInput digitalInput( int index ) const;
    StructDigitalInputExt digitalInputExt( int index ) const;
    int digitalInputCount() const;
};

/////////////////////////////
// Klasse fuer einen Digitalen Eingang

class EntryDigitalInput : public EntryBase
{
public:
    // Konstruktor
    EntryDigitalInput( EntryBase *parent = 0, EntryType type = EntryType::DigitalInput ) : EntryBase( parent, type )
    {
        // Basisvorgabe
        m_indexBase = 0;
    }
    // Destruktor
    ~EntryDigitalInput() override
    {
    }

    ///////////////////////////////////
    // Childs
public:
    EntryBase *insertEntry( EntryType typeID, int pos ) override;
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;
private:
    void checkMembers();

    //////////////////////////////////////
    // Referenzierung zwischen Einträgen
public:
    void convertReferenceToIndex() override;
    void convertIndexToReference() override;
public:
    EntryInputChain *m_refInputChainDigital[2][4];

    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;
    void setDefault( int indexBase );
    int indexBase()
    {
        return m_indexBase;
    }
    void setIndexBase( int indexBase )
    {
        m_indexBase = indexBase;
        setName( boost::str( boost::format( "DI %1% / %2%" ) % (m_indexBase + 1) % (m_indexBase + 2) ) );
    }
private:
    int m_indexBase;
public:
    StructDigitalInput m_digitalInput[2];
    StructDigitalInputExt m_digitalInputExt[2];

};

/////////////////////////////
// Klasse fuer Analoge Eingänge

class EntryAnalogInputs : public EntryBase
{
public:
    // Konstruktor
    EntryAnalogInputs( EntryBase *parent = 0, EntryType type = EntryType::AnalogInputs ) : EntryBase( parent, type )
    {
    }
    // Destruktor
    ~EntryAnalogInputs() override
    {
    }

    ///////////////////////////////////
    // Childs
public:
    EntryBase *insertEntry( EntryType typeID = EntryType::AnalogInput, int pos = -1 ) override;
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;
    void deleteEntry( int pos = -1 ) override;

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;

    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;

    ////////////////////////////////////
    // Schnittstelle des Eintrags
    StructAnalogInput analogInput( int index ) const;
    int analogInputCount() const;
};

/////////////////////////////
// Klasse fuer einen Analogen Eingang

class EntryAnalogInput : public EntryBase
{
public:
    // Konstruktor
    EntryAnalogInput( EntryBase *parent = 0, EntryType type = EntryType::AnalogInput ) : EntryBase( parent, type )
    {
    }
    // Destruktor
    ~EntryAnalogInput() override
    {
    }

    ///////////////////////////////////
    // Childs
public:
    EntryBase *insertEntry( EntryType typeID, int pos ) override;
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;
private:
    void checkMembers() override;

    //////////////////////////////////////
    // Referenzierung zwischen Einträgen
public:
    void convertReferenceToIndex() override;
    void convertIndexToReference() override;
public:
    EntryInputChain *m_refInputChainAnalog[5];

    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;
    void setDefault( int indexBase );
    int indexBase()
    {
        return m_indexBase;
    }
    void setIndexBase( int indexBase )
    {
        m_indexBase = indexBase;
        setName( boost::str( boost::format( "AI %1%" ) % (m_indexBase + 1) ) );
    }

private:
    int m_indexBase;
public:
    StructAnalogInput m_analogInput;
};

/////////////////////////////
// Klasse fuer Timer

class EntryTimers : public EntryBase
{
public:
    // Konstruktor
    EntryTimers( EntryBase *parent = 0, EntryType type = EntryType::Timers ) : EntryBase( parent, type )
    {
    }
    // Destruktor
    ~EntryTimers() override
    {
    }

    ///////////////////////////////////
    // Childs
public:
    EntryBase *insertEntry( EntryType typeID = EntryType::Timer, int pos = -1 ) override;
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;
    void deleteEntry( int pos = -1 ) override;

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;

    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;

    ////////////////////////////////////
    // Schnittstelle des Eintrags
    StructTimer timer( int index ) const;
    int timerCount() const;
};

/////////////////////////////
// Klasse fuer einen Timer

class EntryTimer : public EntryBase
{
public:
    // Konstruktor
    EntryTimer( EntryBase *parent = 0, EntryType type = EntryType::Timer ) : EntryBase( parent, type )
    {
    }
    // Destruktor
    ~EntryTimer() override
    {
    }

    ///////////////////////////////////
    // Childs
public:
    EntryBase *insertEntry( EntryType typeID, int pos ) override;
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;
private:
    void checkMembers() override;

    //////////////////////////////////////
    // Referenzierung zwischen Einträgen
public:
    void convertReferenceToIndex() override;
    void convertIndexToReference() override;
public:
    EntryInputChain *m_refInputChainTimer;

    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;
    void setDefault( int indexBase );
    int indexBase()
    {
        return m_indexBase;
    }
    void setIndexBase( int indexBase )
    {
        m_indexBase = indexBase;
        setName( boost::str( boost::format( "T%1%" ) % (m_indexBase + 1) ) );
    }
private:
    int m_indexBase;
public:
    StructTimer m_timer;
};

/////////////////////////////
// Klasse fuer Systemmeldungen

class EntrySystemMessage;

class EntrySystemMessages : public EntryBase
{
public:
    // Konstruktor
    EntrySystemMessages( EntryBase *parent = 0, EntryType type = EntryType::SystemMessages ) : EntryBase( parent, type )
    {
    }
    // Destruktor
    ~EntrySystemMessages() override
    {
    }

    ///////////////////////////////////
    // Childs
public:
    EntryBase *insertEntry( EntryType typeID = EntryType::SystemMessage, int pos = -1 ) override;
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;
    void deleteEntry( int pos = -1 ) override;

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;

    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;

    ////////////////////////////////////
    // Schnittstelle des Eintrags
    bool exists( EnumSystemMessageType type, UnionSystemMessageId id );
    EntrySystemMessage *add( EnumSystemMessageType type, UnionSystemMessageId id );
    void remove( EnumSystemMessageType type, UnionSystemMessageId id );
    EntrySystemMessage *systemMessage( EnumSystemMessageType type, UnionSystemMessageId id );
    StructSystemMessage systemMessage( int index ) const;
    int systemMessageCount() const;
};

/////////////////////////////
// Klasse fuer eine Systemmeldung

class EntrySystemMessage : public EntryBase
{
public:
    // Konstruktor
    EntrySystemMessage( EntryBase *parent = 0, EntryType type = EntryType::SystemMessage ) : EntryBase( parent, type )
    {
    }
    // Destruktor
    ~EntrySystemMessage() override
    {
    }

    ///////////////////////////////////
    // Childs
public:
    EntryBase *insertEntry( EntryType typeID, int pos ) override;
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;
private:
    void checkMembers() override;

    //////////////////////////////////////
    // Referenzierung zwischen Einträgen
public:
    void convertReferenceToIndex() override;
    void convertIndexToReference() override;
public:
    EntryInputChain *m_refInputChain;

    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;
    void setDefault( int indexBase );
    int indexBase()
    {
        return m_indexBase;
    }
    void setIndexBase( int indexBase )
    {
        m_indexBase = indexBase;
        setName( boost::str( boost::format( "T%1%" ) % (m_indexBase + 1) ) );
    }
private:
    int m_indexBase;
public:
    StructSystemMessage m_systemMessage;
};

/////////////////////////////
// Klasse fuer Eingabeketten der Signalverarbeitung

class EntryEventChains : public EntryBase
{
public:
    // Konstruktor
    EntryEventChains( EntryBase *parent = 0, EntryType type = EntryType::EventChains ) : EntryBase( parent, type )
    {
    }
    // Destruktor
    ~EntryEventChains() override
    {
    }

    ///////////////////////////////////
    // Childs
public:
    EntryBase *insertEntry( EntryType typeID = EntryType::InputChain, int pos = -1 ) override;
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;

    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;
};

/////////////////////////////
// Klasse fuer 

class EntrySignalProcessing : public EntryBase
{
public:
    // Konstruktor
    EntrySignalProcessing( EntryBase *parent = 0, EntryType type = EntryType::SignalProcessing )
        : EntryBase( parent, type )
        , m_systemMessages( new EntrySystemMessages( this ) )
    {
        // Eintrag für Systemmeldungen initalisieren.
        m_systemMessages->setDefault();
    }
    // Konstruktor
    ~EntrySignalProcessing() override
    {
    }

    ///////////////////////////////////
    // Childs
public:
    EntryBase *insertEntry( EntryType typeID, int pos ) override;
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;

    //////////////////////////////////////
    // Referenzierung zwischen Einträgen
public:
    void convertReferenceToIndex() override;
    void convertIndexToReference() override;

public: // Eingabekettenbezuege fuer m_extended.xxx
    std::array<EntryInputChain *, 2> m_refInputChainAdaption;
    EntryInputChain *m_refInputChainGPRS;
    std::array<EntryInputChain *, 5> m_refInputChainGPS;
    std::array<EntryInputChain *, 2> m_refInputChainVibration;

    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;
public:
    StructMachineDefinition m_machineDefinition[6];
    StructVibrationSensor m_vibrationSensor;
private:
    std::unique_ptr<EntrySystemMessages> m_systemMessages;

    ///////////////////////////////////
    // Schnittstelle des Eintrags
public:
    // Gibt Referenz auf ditiale Eingänge zurück.
    EntryDigialInputs &rootDigitalInputs();
    // Gibt Referenz auf analoge Eingänge zurück.
    EntryAnalogInputs &rootAnalogInputs();
    // Gibt Referenz auf timer zurück.
    EntryTimers &rootTimers();
    // Gibt Referenz auf SystemMeldung zurück.
    EntrySystemMessages &rootSystemMessages();
};


/////////////////////////////
// Klasse fuer 

class EntryAccessControl2 : public EntryBase
{
public:
    // Konstruktor
    EntryAccessControl2( EntryBase *parent = 0, EntryType type = EntryType::AccessControl ) : EntryBase( parent, type )
    {
    }
    // Destruktor
    ~EntryAccessControl2() override
    {
    }

    ///////////////////////////////////
    // Childs
    EntryBase *insertEntry( EntryType typeID, int pos ) override;
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;

    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;
protected:
    void setDefaultTable();
    void setDefaultAction2Table();
    void setDefaultInputChain();
public:
    unsigned char m_useExternalReader;
    unsigned char m_entranceMainMenu;
    unsigned char m_accessControlFlag;
    unsigned char m_enableCut;
    unsigned char m_cutAt;
    unsigned char m_cutTo;
    StructTransponder m_transponder;
};

/////////////////////////////
// Klasse fuer 

class EntryEnergieSchemes : public EntryBase
{
public:
    // Konstruktor
    EntryEnergieSchemes( EntryBase *parent = 0, EntryType type = EntryType::EnergieSchemes ) : EntryBase( parent, type )
    {
    }
    // Destruktor
    ~EntryEnergieSchemes() override
    {
    }

    ///////////////////////////////////
    // Childs
    EntryBase *insertEntry( EntryType typeID, int pos ) override;
    // Erzeugt eine Tiefe Kopie des Eintrags.
    EntryBase *cloneEntry() override;

    ///////////////////////////////////
    // Serialisierung der Eintragsdaten
public:
    uint32_t writeDocument( uint32_t pFlash, DocumentStudio &doc ) override;
    uint32_t readDocument( uint32_t pFlash, DocumentStudio &doc ) override;

    ///////////////////////////////////
    // Daten des Eintrags
public:
    void setDefault() override;
    void setDefaultEnergieScheme();
    void verifyEnergieSchemes();

public:
#define PAF_ENABLED                        0x80000000
#define PAF_EXPERT                         0x40000000
#define PAF_INPUT_MANUAL                   0x00000001
#define PAF_INPUT_MANUAL_KEYPAD            0x00000002
#define PAF_INPUT_MANUAL_TRANSPONDER       0x00000004
#define PAF_INPUT_MANUAL_FINGERPRINT       0x00000008
#define PAF_EXTERNAL_ENERGIE               0x00000010
#define PAF_EXTERNAL_ENERGIE_AKKU          0x00000020
#define PAF_EXTERNAL_ENERGIE_AKKU_USB      0x00000040
#define PAF_EXTERNAL_ENERGIE_5V            0x00000080
#define PAF_EXTERNAL_ENERGIE_12V           0x00000100
#define PAF_MOVEMENT                       0x00000200
#define PAF_MOVEMENT_GPS                   0x00000400
#define PAF_MOVEMENT_SENSOR                0x00000800
#define PAF_INPUT_SIGNAL                   0x00001000
#define PAF_INPUT_SIGNAL_TIMER             0x00002000
#define PAF_INPUT_SIGNAL_DIGITAL           0x00004000
#define PAF_INPUT_SIGNAL_ANALOG            0x00008000
#define PAF_MMC                            0x00010000
#define PAF_USB_HOST                       0x00020000
#define PAF_DISPLAY                        0x00040000
    //#define PAF_                            0x000
    PowerSetupNode m_powerSetupNode;
    PowerSetup m_powerSetup[POWER_PROFIL_COUNT];
    PowerEvent m_powerEventTimerA[POWER_EVENT_TIMER_A_MAX];
    PowerEvent m_powerEvent[POWER_EVENT_MAX];
    //PowerEvent m_powerEventTimerB[POWER_EVENT_TIMER_B_MAX];
    PowerMemo m_powerMemo;
};

#endif
